#!/usr/bin/env python3
#
# Script to filter mass mail sendings, and handle quota limitation by periods
#
# Why ?
# Because postfwd is not the correct software to do this (handle period based
# on the first mail sending date)
#
# Guillaume Penaud (gpenaud@easter-eggs.com)
# 05/2015
# 
# exit codes
#   - 0: successful
#   - 1: nothing done
#   - 2: ldap server error
#

"""
Handle quota limitations on postfix
"""
import argparse
import asyncio
import logging
import os
import signal
import sys
import yaml

from logging import handlers

from .api_manager import ApiManager
from .data_manager import DataManager
from .ldap_manager import LdapManager

# Server
# -----------------------------------------------------------------------------


class Server(object):

    config_filepath = "/etc/postquota/configuration.yaml"

    logger = None
    log_handler = None

    config = {}
    args   = {}

    def __init__(self):
        # mixed variables declaration
        self.stopped = False
        self.attrs = {}

# Load configuration object
# -----------------------------------------------------------------------------
    def load_config(self):
        config_filepath = self.args.get("configuration_file")

        if config_filepath is None:
            config_filepath = self.config_filepath

        # handle configuration
        try:
            stream = open(config_filepath, "r")
        except IOError:
            sys.stderr.write('[ERROR] Impossible to load configuration file\n')
            sys.exit(1)

        # set-up server configuration
        self.config = yaml.load(stream, Loader=yaml.SafeLoader)

        # load logger here to be able to change configuration upon it
        self.load_logger()

        self.hostname = self.config.get("global", {}).get("policy_host", "127.0.0.1")
        self.port = self.config.get("global", {}).get("policy_port", 9001)

        # handle periods
        self.periods = {}

        available_periods = ["hourly", "daily", "weekly", "monthly", "yearly"]
        periods = self.config.get("period")

        if not all(period in available_periods for period in periods.keys()):
            self.logger.error('One of the period set is unauthorized')
            sys.exit(2)

        sorted_periods = list(periods.items())
        sorted_periods.sort(key=lambda v: v[1]["nb_seconds"])
        sorted_periods.reverse()

        for label, data in sorted_periods:
            self.periods.setdefault(label, data)


# Set-up custom logger
# -----------------------------------------------------------------------------

    def load_logger(self):
        self.logger = logging.getLogger('postquota')

        if self.log_handler is not None:
            self.logger.removeHandler(self.log_handler)

        log_level = self.config.get("global", {}).get("log_level", "INFO")
        self.logger.setLevel(log_level)

        log_target = self.config.get("global", {}).get("log_target", "syslog")
        if log_target == "stderr":
            self.log_handler = logging.StreamHandler(sys.stderr)
        elif log_target == "stdout":
            self.log_handler = logging.StreamHandler(sys.stdout)
        else:
            mail_facility = logging.handlers.SysLogHandler.LOG_MAIL
            self.log_handler = logging.handlers.SysLogHandler(facility=mail_facility, address='/dev/log')

        # to have a nice log format
        formatter = logging.Formatter("%(name)s: %(levelname)s: %(message)s")
        if self.logger.getEffectiveLevel() <= logging.DEBUG:
            formatter = logging.Formatter("%(levelname)-8.8s provider: %(name)-10.12s %(asctime)-19.19s [%(module)s.py l.%(lineno)s %(funcName)s()] %(message)s")
        self.log_handler.setFormatter(formatter)
        self.logger.addHandler(self.log_handler)


# Handle client connections
# -----------------------------------------------------------------------------

    async def handle_connection(self, reader, writer):
        self.logger.debug("Handle socket connection")
        while True:
            try:
                req = (await reader.readuntil(separator=b"\n\n")).decode("utf8").strip()
                self.logger.debug(f"Received request:\n{req}")

                attrs = self.parse_request(req)
                if "request" not in attrs:
                    self.logger.error("Missing key 'request'.")

                res = "action=%s\n\n" % self.smtpd_access_policy(attrs)
                writer.write(res.encode("utf-8") + b"\n\n")
                await writer.drain()
                self.logger.debug(f"Sent response:\n{res}")
            except asyncio.exceptions.IncompleteReadError as e:
                last_data_received = e.partial.decode("utf-8").strip()
                if last_data_received:
                    self.logger.error(f"Remote side closed the connection. Last data received: '{last_data_received}'")
                else:
                    self.logger.debug(f"Remote side closed the connection. Last data received: '{last_data_received}'")
                writer.close()
                break
            except Exception as e:
                self.logger.exception("Unexpected error while handling request", exc_info=True)
                break


# Handle user-defined signals
# -----------------------------------------------------------------------------

    def handle_signal(self, signum, frame):
        if signum == signal.SIGUSR1:
            self.logger.debug("Signal %s triggered - flushing cache" % signum)
            self.datamanager.flush()
        elif signum == signal.SIGUSR2:
            self.logger.debug("Signal %s triggered - reloading configuration" % signum)
            self.load_config()
        elif signum == signal.SIGHUP:
            self.logger.debug("Signal %s triggered - reloading cache" % signum)
            self.datamanager.load()
        elif signum == signal.SIGINT or signum == signal.SIGTERM:
            self.logger.debug("Signal %s triggered - flush then stop daemon" % signum)
            self.datamanager.flush()
            self.stop()
        else:
            self.logger.debug("Unknown signal %s triggered - pass" % signum)


# Entrance of the request, apply regex on input stream
# -----------------------------------------------------------------------------

    def parse_request(self, data):
        attrs = {}
        for line in data.split("\n"):
            line = line.strip()
            if "=" not in line:
                self.logger.warning(f"Ignoring invalid line: '{line}'")
                continue
            key, val = line.split("=", 1)
            attrs[key.strip()] = val.strip()
        return attrs


# Handle treatments relative to policy access
# -----------------------------------------------------------------------------

    def smtpd_access_policy(self, attrs):
        self.logger.debug("Handle mail policy")

        if "sasl_username" not in attrs:
            self.logger.warning("Attr 'sasl_username' is missing. Ignoring.")
            return ""
        queue_id = attrs.get("queue_id", "?UNKNOWN-QUEUE-ID?")

        action = "DUNNO"
        messages = []
        increment = True

        user_with_fqdn = attrs.get("sasl_username")
        # we need to clean username here, to remove the domain part
        user = user_with_fqdn.split("@")[0]

        if user is None:
            self.logger.info("No user, so the mail shall pass" % user)
            return ""

        for period in self.periods:
            count = self.datamanager.retrieve(user, period, "count")
            try:
                limit = self.ldapmanager.get_limit(user, period)
            except Exception as e:
                self.logger.warning(f"Unspecified limit for user={user}, period={period}. Considering 'unlimited'...")
                continue

            if int(limit) == -2:
                #return "REJECT Service temporarily unavailable"
                continue
            # display "unlimited" instead of "-1"
            elif int(limit) == -1:
                limit_str = "unlimited"
            else:
                limit_str = limit

            # limit reached
            if ((int(limit) > -1) and (int(count) >= int(limit))):
                increment = False
                action = f"REJECT {period} quota exhausted"
                messages.append("[%s: NO (%s/%s)]" % (period, count, limit_str))
                self.logger.debug("Data for period: %s -- count:%s & limit:%s" % (period, count, limit_str))
                break
            else:
                messages.append("[%s: YES (%s/%s)]" % (period, count, limit_str))
                self.logger.debug("Data for period: %s -- count:%s & limit:%s" % (period, count, limit_str))

        # if an increment has been done previously, we need to try to save cache
        if increment:
            self.datamanager.increment(user)
            self.logger.info(f"Mail {queue_id} has been accepted -- %s" % " - ".join(messages))
        else:
            self.logger.info(f"Mail {queue_id} has been rejected -- %s" % " - ".join(messages))

        self.datamanager.save()

        return action + f" -- {queue_id} sasl_username='{user}' " + " - ".join(messages)

# Start server
# -----------------------------------------------------------------------------

    async def start(self):

        self.logger.debug("Starting, please wait...")

        # load the managers
        self.datamanager = DataManager(self)
        self.ldapmanager = LdapManager(self)
        self.api_manager = ApiManager(self)

        self.pol_server = await asyncio.start_server(
            self.handle_connection,
            self.hostname,
            self.port,
        )
        self.api_server = await asyncio.start_server(
            self.api_manager.handle_connection,
            self.config.get("api", {}).get("api_host", "0.0.0.0"),
            self.config.get("api", {}).get("api_port", 9002),
        )

        await asyncio.gather(
            asyncio.ensure_future(self.pol_server.serve_forever()),
            asyncio.ensure_future(self.api_server.serve_forever()),
            return_exceptions=True
        )


# Stop server
# -----------------------------------------------------------------------------

    def stop(self):
        # check 
        states = self.check_connections()

        self.logger.debug("Stopping API server, please wait...")
        self.api_manager.stop()
        self.logger.debug("The API server is stopped")

        self.logger.debug("Stopping policy server, please wait...")
        # TODO
        self.logger.debug("Policy server is closed")

        self.stopped = True


# Main method
# -----------------------------------------------------------------------------

if __name__ == "__main__":
    server = Server()

    # handle cli arguments
    parser = argparse.ArgumentParser(prog="postquota")

    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-d", "--daemonize", action="store_true")
    parser.add_argument(
        "-c", "--configuration-file", default="/etc/postquota/configuration.yaml"
    )

    args = parser.parse_args()
    server.args = args.__dict__

    # load and propagate configuration
    server.load_config()

    # run server
    asyncio.run(server.start())
