# /usr/bin/env python
#
# Handle some internal functions through HTTP server, and rest API
#
# Why ?
# Because this is far more powerful and beautiful than a telnet or
# a TCP socket
#
# Guillaume Penaud (gpenaud@easter-eggs.com)
# 05/2015

import json
from urllib.parse import urlparse
import logging
from pprint import pformat

from urllib.parse import urlparse, parse_qs


class ApiManager():

    def __init__(self, server):
        self.server = server
        self.config = server.config
        self.logger = server.logger

        # handle commands
        self.api_commands = {
            "load": self.handle_load_command,
            "dump": self.handle_dump_command,
            "flush": self.handle_flush_command,
            "get_limits": self.handle_get_limits_command,
            "get_periods": self.handle_get_periods_command,
            "reload_conf": self.handle_reload_configuration_command,
        }

    async def handle_connection(self, reader, writer):
        r = (await reader.readuntil(b"\r\n\r\n"))
        r = r.replace(b"\r\n", b"\n")
        req = r.decode("utf8").strip()
        self.logger.debug(f"Read:\n{req}")
        if req.startswith("GET /stats"):
            res = pformat(self.server.datamanager.data)
            #res = self.server.datamanager.dump()
            self.logger.debug(f"Received request:\n{req}")
            writer.write(res.encode("utf-8") + b"\n")
            await writer.drain()
            self.logger.debug(f"Sent response:\n{res}")
            writer.close()
        elif req.startswith("GET /?cmd="):
            url = req.split(" ")[1]
            query_params = parse_qs(urlparse(url).query)
            cmd_name = query_params.get("cmd")
            if not isinstance(cmd_name, list):
                cmd = None
            else:
                cmd = self.api_commands.get(cmd_name[0])
            if cmd is None:
                self.logger.warning(f"Cannot parse HTTP request: '{req}'")
                res = "HTTP/1.1 400\r\n\r\n"
                res += "Bad request"
            else:
                query_params.pop("cmd")
                res = "HTTP/1.1 200\r\n\r\n"
                try:
                    data = await cmd(args=query_params) or "OK"
                    res += str(data)
                except Exception as e:
                    if self.logger.getEffectiveLevel() <= loggin.DEBUG:
                        self.logger.warning("Unhandled error.", exc_info=True)
                    res = "HTTP/1.1 500\r\n\r\n"
                    res += str(e)
            writer.write(res.encode("utf-8") + b"\n")
            await writer.drain()
            writer.close()
        else:
            self.logger.warning(f"Received request:\n{req}")
            self.logger.warning("Nothing done.")
            writer.close()


# API methods
# --------------------------------------------------------------------------
    
    # load cache file content to RAM memory
    async def handle_load_command(self, args):
        return self.server.datamanager.load()

    async def handle_dump_command(self, args):
        # Those are already up-to-date data
        return self.server.datamanager.dump()

    async def handle_flush_command(self, args):
        self.server.datamanager.flush()

    async def handle_reload_configuration_command(self, args):
        self.server.load_config()

    async def __get_periods(self):
        return [p for p in self.server.periods.keys()]

    async def handle_get_periods_command(self, args):
        return ",".join(await self.__get_periods())

    async def handle_get_limits_command(self, args):
        user = args.get("user", [None])[0]
        limits = {}

        if user is None:
            raise Exception("Unable to retrieve limit, missing user")

        for period in await self.__get_periods():
            limits.setdefault(period, self.server.ldapmanager.get_limit_from_ldap(user, period))

        return json.dumps(limits)
