#!/usr/bin/env python3
#
# Cache for data
#
# Why ?
# Because cache must be accessed by multiple users simultanously
#
# Guillaume Penaud
# 05/2015
#

import json
import os
import os.path
import sys
import time

from datetime import datetime, timedelta
from .ldap_manager import LdapManager


class DataManager():

    def __init__(self, server):
        self.logger = server.logger

        self.config = server.config
        self.periods = server.periods

        self.ldapmanager = LdapManager(server)

        self.last_save = time.time()
        self.save_every = int(self.config.get("data", {}).get("save_every", 60))

        # handle cache file
        cache_file = self.config.get("data", {}).get("file")

        if cache_file is None:
            self.logger.warning('No path defined for cache file')
        elif not os.path.isfile(cache_file):
            os.makedirs(os.path.dirname(cache_file), exist_ok=True)
            with open(cache_file, "w") as f:
                f.write("{}")
            self.logger.warning(f"Initialized an empty cache file: {cache_file}")
        elif not os.access(cache_file, os.R_OK + os.W_OK):
            raise Exception(f"Insufficient permissions on cache file '{cache_file}'")
        
        self.cache_file = cache_file

        # set default data
        self.default_field_values = {"count": 0, "limit": 0}
        self.data = {}

        try:
            self.load()
        except Exception as e:
            self.logger.exception(e.message)


# Determine if a period has been exceeded since last mail send
# -----------------------------------------------------------------------------

    def is_period_exceeded(self, user, period):
        self.logger.debug('Check for period ttl exceeded for user %s at period %s' % (user, period))
        now_ts = datetime.now().timestamp()
        until_ts = self.data.get(user, {}).get(period, {}).get("until")

        if until_ts is None:
            return True

        if now_ts >= until_ts:
            return True

        return False


# Reset cache count for one user on a period
# -----------------------------------------------------------------------------

    def reset_count_for(self, user, period):
        self.logger.debug('Reset count of user %s for the period %s' % (user, period))
        now = datetime.now()

        determine_limit = {
            "hourly": self.get_hourly_limit,
            "daily": self.get_daily_limit,
            "weekly": self.get_weekly_limit,
            "monthly": self.get_monthly_limit,
            "yearly": self.get_yearly_limit,
        }

        until = determine_limit[period](now)

        data = self.data.setdefault(user, {}).setdefault(period, {})
        data["count"] = 0
        data["until"] = int(until.timestamp())


# Periods limit calculation methods
# -----------------------------------------------------------------------------

    def get_hourly_limit(self, now):
        return datetime(now.year, now.month, now.day, now.hour, 0, 0) + timedelta(hours=1)

    def get_daily_limit(self, now):
        return datetime(now.year, now.month, now.day, 0, 0, 0) + timedelta(days=1)

    def get_weekly_limit(self, now):
        dt = datetime(now.year, now.month, now.day, 0, 0, 0) + timedelta(days=7)
        return dt - timedelta(days=dt.weekday())

    def get_monthly_limit(self, now):
        dt = datetime(now.year, now.month, 1, 0, 0, 0) + timedelta(days=32)
        return dt.replace(day=1)

    def get_yearly_limit(self, now):
        return datetime((now.year + 1), 1, 1, 0, 0, 0)


# Get a field value of the main cache
# -----------------------------------------------------------------------------

    def retrieve(self, user, period, field):
        if self.is_period_exceeded(user, period):
            self.reset_count_for(user, period)

        field_value = self.data.get(user, {}).get(period, {}).get(field, self.default_field_values[field])

        return field_value


# Increment count for a user / period
# -----------------------------------------------------------------------------

    def increment(self, user):
        self.logger.debug('Increment cache for %s' % user)

        for period in self.periods:
            data = self.data.setdefault(user, {}).setdefault(period, {})
            count = data.setdefault("count", 0)
            data["count"] = count + 1


# Populate cache with a file
# -----------------------------------------------------------------------------

    def load(self):
        self.logger.debug('Load from cache file')
        try:
            self.data_handler = open(self.cache_file, "r")
            serialized_cache = self.data_handler.read()

            self.data = json.loads(serialized_cache)
        except ValueError:
            self.logger.warning('Cache is empty')
            self.data = {}
        except IOError:
            self.logger.error('Error while accessing cache file')
            self.data = {}


# Determine if this is time to flush cache or not
# -----------------------------------------------------------------------------

    def save(self):
        self.logger.debug('Save to cache file')

        if (self.last_save is None) or (self.last_save < (time.time() - self.save_every)):
            self.flush()

# Dump cache into a file
# -----------------------------------------------------------------------------

    def flush(self):
        self.logger.debug('Flush cache content to cache file')

        try:
            serialized_cache = json.dumps(self.data)
            cache = open(self.cache_file, "w")
            cache.write(serialized_cache)
            self.last_save = time.time()
        except IOError:
            self.logger.error('Can not open cache file, please check')
            sys.exit(7)


# Determine if this is time to flush cache or not
# -----------------------------------------------------------------------------

    def dump(self):
        self.logger.debug('Dump cache content to STDOUT')
        return json.dumps(self.data)
