#!/usr/bin/env python3
#
# LDAP requester with a cache implementation
#
# Why ?
# Because the policy server requires LDAP data to determine limitations
# by user, and could overload LDAP servers
#
# Guillaume Penaud (gpenaud@easter-eggs.com)
# 05/2015
#

"""
Handle LDAP requests
"""

# LDAP manager
# -----------------------------------------------------------------------------

import ldap
import sys
import time


class LdapManager():
    def __init__(self, server):
        self.retry_count = 0
        self.server = server
        self.logger = server.logger

        self.config = server.config
        self.periods = server.periods

        self.cache = {}


# Get cache information
# -----------------------------------------------------------------------------

    def get_limit(self, user, period):
        cache_key = user + "-" + period
        if cache_key not in self.cache:
            self.get_limit_from_ldap(user, period)
        cache_entry = self.cache.get(cache_key)
        expiration =  cache_entry.get("expiration")
        if expiration is None or expiration <= time.time():
            self.logger.debug(f"Cache entry expired for user {user}, regenerating...")
            self.get_limit_from_ldap(user, period)
        limit = self.get_limit_from_cache(user, period)
        return limit


# Tries to retrieve a value from cache entry - fallback to ldap request if an
# exception is triggered
# -----------------------------------------------------------------------------

    def get_limit_from_cache(self, user, period):
        self.logger.debug("Retrieve limit from cache for user %s at period %s" % (user, period))
        limit = None

        cache_key = user + "-" + period
        cache_entry = self.cache.get(cache_key)

        if cache_entry is None:
            raise ValueError("No cache entry for cache_key: %s-%s" % (user, period))

        limit = cache_entry.get("limit")
        expiration = cache_entry.get("expiration")

        if limit is None:
            raise ValueError("Limit is unknown for user %s and period %s" % (user, period))

        if expiration is None or expiration <= time.time():
            raise ValueError("Cache entry expirated for user %s and period %s, regeneration..." % (user, period))

        return limit


# Check ldap connection, and initialize if needed
# -----------------------------------------------------------------------------

    def open_ldap_connection(self):
        self.logger.debug("Open LDAP connection")
        uri = self.config.get("ldap", {}).get("uri", {})

        try:
            self.ldap_connection = ldap.initialize(uri)
            self.ldap_connection.set_option(ldap.OPT_NETWORK_TIMEOUT, 10.0)

            bind_dn = self.config.get("ldap", {}).get("bind_dn", {})
            bind_pw = self.config.get("ldap", {}).get("bind_pw", {})

            if bind_dn is not None and bind_pw is not None:
                self.ldap_connection.simple_bind_s(bind_dn, bind_pw)

        except ldap.LDAPError as e:
            self.logger.error(f"Failed to establish connection to LDAP (attempt {self.retry_count+1}/5). Cause: {str(e)}")
            if self.retry_count < 4:
                self.retry_count = self.retry_count + 1
                time.sleep(5)
                self.open_ldap_connection()
            else:
                self.logger.critical("Stopping after %s LDAP connection attempts" % self.retry_count)
                self.retry_count = 0


# Refresh cache by ldap remote data
# -----------------------------------------------------------------------------
    def get_limit_from_ldap(self, user, period_to_retrieve):
        self.logger.debug("Retrieve limits from LDAP for user %s" % user)
        try:
            self.open_ldap_connection()
            filters = self.config.get("ldap", {}).get("filter", "(uid=%s)") % user
            fields = [v_period["field"] for v_period in self.periods.values()]
            fields.append("uid")

            search_base = self.config.get("ldap", {}).get("search_base")

            result_id = self.ldap_connection.search(search_base, ldap.SCOPE_SUBTREE, filters, fields)
            result_type, result_data = self.ldap_connection.result(result_id, 2)

            if len(result_data) > 1:
                raise ValueError("LDAP search has returned ambiguous value, please change filter criteria")

            if result_type == ldap.RES_SEARCH_ENTRY:
                ldap_data = result_data[0][1]

                for period, period_data in self.periods.items():
                    field = period_data.get("field", {})
                    limits = ldap_data.get(field)

                    if limits is None:
                        limit = '-1'
                    else:
                        limit = limits[0]

                    if isinstance(limit, bytes):
                        limit = limit.decode("utf-8")

                    self.update_limit_in_cache(user, period, limit)

            return self.get_limit_from_cache(user, period_to_retrieve)
        except ldap.LDAPError as e:
            self.logger.error("Failed to fetch quota information from LDAP. Cause: " + str(e))
            return -2


# Refresh cache by retrieving ldap remote data
# -----------------------------------------------------------------------------

    def update_limit_in_cache(self, user, period, limit):
        self.logger.debug("Update limit (%s) for user %s at period %s" % (limit, user, period))
        cache_key = user + "-" + period

        expiration = self.config.get("ldap", {}).get("cache_expiration_delay", 60)
        now = time.time()
        delay = now + int(expiration)

        # initialize hash if he does not already exists
        cache = self.cache.setdefault(cache_key, {})

        cache["limit"] = limit
        cache["expiration"] = delay
        cache["last_update"] = now

        return cache

# End of file
# -----------------------------------------------------------------------------
